// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2010 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************
//
#include "unittest.h"
#include "../resources.h"
#include "../../shared/fileHandling.h"
#include "../multithreading.h"
#include "../../algorithm.h"
#include "../../shared/globalFunctions.h"
#include "../../ui/util.h"
#include "../filter.h"
#include "../../shared/stringConv.h"

using namespace FreeFileSync;



TEST(Miscellaneous, RemoveRowsFromVector)
{
    std::vector<int> test; //0 1 2 3 4 5 6
    for (int i = 0; i < 7; ++i)
        test.push_back(i);

    std::set<unsigned int> rowsToRemove;
    rowsToRemove.insert(-41); //invalid
    rowsToRemove.insert(1);
    rowsToRemove.insert(2);
    rowsToRemove.insert(5);
    rowsToRemove.insert(15);  //invalid
    rowsToRemove.insert(100); //invalid

    globalFunctions::removeRowsFromVector(rowsToRemove, test);
    //0 3 4 6

    ASSERT_EQ(unsigned(4), test.size());
    ASSERT_EQ(0, test[0]);
    ASSERT_EQ(3, test[1]);
    ASSERT_EQ(4, test[2]);
    ASSERT_EQ(6, test[3]);
}


TEST(Formatting, DirectoryNames)
{
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(FreeFileSync::getFormattedDirectoryName(DefaultStr("  c:\\test "))), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(FreeFileSync::getFormattedDirectoryName(DefaultStr("  c:\\test\\ "))), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(FreeFileSync::getFormattedDirectoryName(DefaultStr("  \\abc\\ "))), wxT("C:\\abc\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(FreeFileSync::getFormattedDirectoryName(DefaultStr("  \\a\\b\\c "))), wxT("C:\\a\\b\\c\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(FreeFileSync::getFormattedDirectoryName(DefaultStr("  \\a\\b\\c d  "))), wxT("C:\\a\\b\\c d\\"));
    ASSERT_PRED_FORMAT2(assertEqual, zToWx(FreeFileSync::getFormattedDirectoryName(DefaultStr("  \\\\a b/b\\c d  "))), wxT("\\\\a b\\b\\c d\\"));
}


//predicate-formatter for asserting that two vector<Zstring> are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const std::vector<Zstring>& a, const std::vector<Zstring>& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = a.begin(); i != a.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " and " << b_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = b.begin(); i != b.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " not equal!";

    return testing::AssertionFailure(msg);
}

std::vector<Zstring> compoundStringToFilter(const Zstring& filterString);

TEST(Formatting, FilterString)
{
    std::vector<Zstring> filterActual = compoundStringToFilter(DefaultStr("   A\n   b;; c;\n\n;\n d    ;E    f  \n;& ;g\\"));
    std::vector<Zstring> filterExpected;
    filterExpected.push_back(DefaultStr("A"));
    filterExpected.push_back(DefaultStr("b"));
    filterExpected.push_back(DefaultStr("c"));
    filterExpected.push_back(DefaultStr("d"));
    filterExpected.push_back(DefaultStr("E    f"));
    filterExpected.push_back(DefaultStr("&"));
    filterExpected.push_back(DefaultStr("g\\"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(DefaultStr("   ; ;\n a"));
    filterExpected.clear();
    filterExpected.push_back(DefaultStr("a"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(DefaultStr("   ; ;\n       ;\n\n"));
    filterExpected.clear();
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);
}


TEST(Formatting, FileSize)
{
    wxString sizeActual = FreeFileSync::formatFilesizeToShortString(807.0);
    wxString sizeExpected = wxString(wxT("807")) + _(" Byte");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(315621376.0);
    sizeExpected = wxString::Format(wxT("%i"), 301) + _(" MB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(48954.0);
    sizeExpected = wxString(wxT("47.8")) + _(" kB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(35032904.0);
    sizeExpected = wxString(wxT("33.4")) + _(" MB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(3084582912.0);
    sizeExpected = wxString(wxT("2.87")) + _(" GB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(1001.0);
    sizeExpected = wxString(wxT("0.97")) + _(" kB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);
}


TEST(Filtering, Standard)
{
    FreeFileSync::NameFilter filterIncl(DefaultStr("*abc?5*2  ;    abc?3;   name\\; test\\subdir "), DefaultStr(""));

    ASSERT_TRUE(filterIncl.passFileFilter( DefaultStr("zzzzabcz5zzzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(DefaultStr("zzzzabczz5zzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(DefaultStr("zzzzabcz5zzzzz")));

    ASSERT_TRUE(filterIncl.passFileFilter(DefaultStr("abcz3")));
    ASSERT_FALSE(filterIncl.passFileFilter(DefaultStr("abcz")));

    bool continueTrav = true;
    ASSERT_TRUE(filterIncl.passDirFilter(DefaultStr("name"), &continueTrav));

    ASSERT_FALSE(filterIncl.passDirFilter(DefaultStr("test"), &continueTrav));
    ASSERT_TRUE(continueTrav); //but continue traversing, because subdir might match filter
//------------------------------------------------------------------------------------------

    FreeFileSync::NameFilter filterExcl(DefaultStr("*"), DefaultStr("asdf?5 ;;; name\\*"));

    ASSERT_FALSE(filterExcl.passFileFilter(DefaultStr("asdfz5")));
    ASSERT_TRUE(filterExcl.passFileFilter(DefaultStr("asdfz")));

    ASSERT_FALSE(filterExcl.passDirFilter(DefaultStr("name"), &continueTrav));
    ASSERT_FALSE(continueTrav);
}


//handle execution of a method while updating the UI
class UpdateDummy : public UpdateWhileExecuting
{
private:
    void longRunner() {} //virtual impl.
};


TEST(Miscellaneous, Standard)
{
    //test if instantiation and deletion of multithreading class works correctly (without hang)
    for (int i = 0; i < 50; ++i)
    {
        UpdateDummy d;
    }

    //test if multiple executions work (without hang)
    UpdateDummy d;
    DummyStatusUpdater dummy;
    for (int i = 0; i < 1000; ++i)
    {
        d.waitUntilReady();
        d.execute(&dummy);
    }
}


TEST(Zstring, AfterBefore)
{
    ASSERT_EQ(Zstring(DefaultStr("a/dsfsadf/abc")).AfterLast(DefaultChar('/')), Zstring(DefaultStr("abc")));
    ASSERT_EQ(Zstring(DefaultStr("a/dsfsadf/abc")).BeforeLast(DefaultChar('/')), Zstring(DefaultStr("a/dsfsadf")));
    ASSERT_EQ(Zstring(DefaultStr("a/dsfsadf/abc")).AfterFirst(DefaultChar('/')), Zstring(DefaultStr("dsfsadf/abc")));
    ASSERT_EQ(Zstring(DefaultStr("a/dsfsadf/abc")).BeforeFirst(DefaultChar('/')), Zstring(DefaultStr("a")));
}


TEST(Zstring, Resize)
{
    Zstring testStr(DefaultStr("abcdefgh"));
    testStr.resize(3);
    ASSERT_EQ(testStr, Zstring(DefaultStr("abc")));

    Zstring testStr2(DefaultStr("abcdefgh"));
    testStr2.resize(10); //add 0-characters!
    ASSERT_EQ(testStr2, Zstring(DefaultStr("abcdefgh\0\0"), 10));

    Zstring testStr3(DefaultStr("abcdefgh"));
    testStr3.resize(3, DefaultChar('N'));
    ASSERT_EQ(testStr3, Zstring(DefaultStr("abc")));

    Zstring testStr4(DefaultStr("abcdefgh"));
    testStr4.resize(10, DefaultChar('N')); //add 'N' characters!
    ASSERT_EQ(testStr4, Zstring(DefaultStr("abcdefghNN")));
}
